/* written by John Doug Reynolds, June 2002 */

#include "calib.h"
#include "param.h"
#include <CalServer.h>
#include <Convolver.h>
#include <errstr.h>
#include <algorithm>
#include <vector>

using namespace scanningLib;
using namespace std;

static vector<float>& wrap( vector<float>&, const Spectrum& );
static void shift( vector<float>&, const Spectrum&, int offset );
static void rebin( struct Spectrum &vec, float min, float max, float val );

static double calBeta( double phi /* arcmin, dispersion off-axis angle */,
		       int order, double energy /* eV */ );

static Spectrum customFigureDistribution( double betaBinWidth,
					  int order, double energy /* eV */ );

static CcdQuantumEfficiencyCurve	ccdqeCurve[9];
static EffectiveAreaDataServer*		caldsEA;
static EffectiveAreaCurve*		eaCurve[6];
static GeometryDataServer*		caldsGM;
static GratingDataServer*		caldsGR;
static PsfDataServer*			caldsPS;
static GridSpan				betaGrid;
static double				theta, phi;
//AI: Wavelength scale. Start
static GridSpan betaChannelGrid;
static GridSpan betaORIChannelGrid;
//AI: Wavelength scale. End
vector<AngularDistBin> customFigure;
vector<double> vignetting;

/* --- LSF components ------------------------------------------------------ */

#if 1
void prfBroad( vector<float> &rf,
	       float eLow, float eHigh, float alpha, int order, string energyBinning )
{
  //AI: Wavelength scale. Start
  double energy = 1000. * (eLow + eHigh) / 2.;
  double beta_m = caldsGR->beta(alpha,order,energy);

  //int curveHead = betaGrid.cell(beta_m) - betaGrid.min;
  //AI: lambda binning
  double lambda = PhysicalConstants::eVtimesAo*(-order)/energy;
  int curveHead;
  double dBeta;
  if (energyBinning == "lambda") {
	  double energy_m=PhysicalConstants::eVtimesAo*(-order)/energy; //Lambda space


	  //Get the cell corresponding with this lambda value
	  //cout << "energy_m " << energy_m << endl;
	  double lambdaMin = betaChannelGrid[betaChannelGrid.tick(energy_m)]-(betaChannelGrid.stride/2.);
	  double lambdaMax = betaChannelGrid[betaChannelGrid.tick(energy_m)]+(betaChannelGrid.stride/2.);
	  //cout << "lambdaMin " << lambdaMin << " lambdaMax "<< lambdaMax << endl;
	  double betaMin = caldsGR->lambda2beta(alpha,-1,lambdaMin);
	  double betaMax = caldsGR->lambda2beta(alpha,-1,lambdaMax);
	  dBeta = std::abs(betaMax-betaMin);

	  curveHead = betaChannelGrid.cell(lambda) - betaChannelGrid.min; //AI: Wavelength scale. Start
  }
  else
  {
	  dBeta =  betaChannelGrid.stride;
	  curveHead = betaChannelGrid.cell(beta_m) - betaChannelGrid.min; //AI: Wavelength scale. Start
  }
  //AI: lambda binning

  int right = max( (int) betaChannelGrid.length() - curveHead - 1, 0 ); //AI: Wavelength scale. Start
  int left = max( curveHead, 0 );

  Spectrum
    dist=caldsGR->scatteringDistribution(left, right, dBeta,
  					 alpha, order, energy, false, true);  //AI: Wavelength scale. Start

  shift( rf, dist, curveHead - left );

  float effective_area = eaCurve[-order]->area(1000.*eLow,1000.*eHigh);
  for ( unsigned i = 0; i < betaChannelGrid.length(); ++i ) rf[i] *= effective_area; //AI: Wavelength scale. Start

}
#else
void prfBroad( vector<float> &rf,
	       float eLow, float eHigh, float alpha, int order, string energyBinning )
{
  double energy = 1000 * (eLow + eHigh) / 2;
  double beta_m = caldsGR->beta(alpha,order,energy);

  Spectrum
    dist=caldsGR->scatteringDistribution(100, betaGrid.stride,
					 alpha, order, energy, false, true);

  shift( rf, dist, betaGrid.cell(beta_m)-betaGrid.min-(int)dist.size()/2 );

  float effective_area = eaCurve[-order]->area(1000*eLow,1000*eHigh);
  for ( unsigned i = 0; i < betaChannelGrid.length(); ++i ) rf[i] *= effective_area;
}
#endif

void prfNarrow( vector<float> &rf,
		float eLow, float eHigh, float alpha, int order,
		int fftdim, bool withmirrorpsf, bool withangdist, string energyBinning)
{

	//AI: Lambda binning
	double energy = 1000. * (eLow + eHigh) / 2.;

	double beta_m = caldsGR->beta(alpha,order,energy);
	double dBeta;
	double energy_m;
  if (energyBinning == "lambda")
  {

	  energy_m=PhysicalConstants::eVtimesAo*(-order)/energy; //Lambda space

	  //Get the cell corresponding with this lambda value
	  double lambdaMin = betaChannelGrid[betaChannelGrid.tick(energy_m)]-(betaChannelGrid.stride/2.);
	  double lambdaMax = betaChannelGrid[betaChannelGrid.tick(energy_m)]+(betaChannelGrid.stride/2.);
	  double betaMin = caldsGR->lambda2beta(alpha,-1,lambdaMin);
	  double betaMax = caldsGR->lambda2beta(alpha,-1,lambdaMax);
	  dBeta = std::abs(betaMax-betaMin);


  }
  else
  {
	  dBeta = betaChannelGrid.stride;
	  energy_m = beta_m;
  }
  //AI: Lambda binning



  //AI: Lambda binning
  struct Spectrum dist;

  if ( betaChannelGrid.contains( betaChannelGrid.cell(energy_m) ) )
    {

      list<Spectrum> kernels;
      double cm = std::sin(alpha) / std::sin(beta_m); // chromatic magnification
      double nhwhm;

      // small angle scattering

      if ( fftdim == 1 ) nhwhm = 1.5;
      else if ( fftdim == 2 ) nhwhm = 3;
      else nhwhm = 4;

      //cout << "dBeta " << dBeta << endl;
      //cout << "alpha " << alpha << endl;
      //cout << "order " << order << endl;
      //cout << "energy " << energy << endl;
      dist = caldsGR->scatteringDistribution(nhwhm,dBeta,
					     alpha,order,energy,true,false);

      DPair fSmallLarge = caldsGR->scatteringDistributionContribs();

      dist[dist.size()/2] += (1 - fSmallLarge.first)*(1 - fSmallLarge.second);

      kernels.push_back( dist );

      // mirror point spread contribution

      if ( withmirrorpsf ) {
    	  if ( fftdim == 1 ) nhwhm = 36;
    	  else if ( fftdim == 2 ) nhwhm = 72;
    	  else if ( fftdim == 3 ) nhwhm = 157;
    	  else if ( fftdim == 4 ) nhwhm = 340;
    	  else nhwhm = 706;

    	  kernels.push_back
    	  	  ( caldsPS->dispersionFigureDistribution(nhwhm,dBeta,cm) );
      }

      // custom angular distribution

      if ( withangdist )
      {
    	  kernels.push_back
    	  ( customFigureDistribution(dBeta,order,energy) );
      }

      // grating bow induced misalignment
      kernels.push_back
      ( caldsGR->bowingFigureDistribution(8,dBeta,cm) );

      // geometric defocus approximation
      kernels.push_back
      ( caldsGM->lsfDefocusDistribution(1,dBeta,theta,phi,order,energy) );

      // finite energy band contribution

      double dLambda = hc/eLow - hc/eHigh;
      kernels.push_back
     ( caldsGR->broadeningDistribution(dBeta,dLambda,beta_m,order) );

      // grating misalignment
      double ico = betaChannelGrid.tick(energy_m) - betaChannelGrid.cell(energy_m);
      dist = caldsGR->misalignmentFigureDistribution(8,dBeta,cm,ico);

      // convolve grating misalignment against the accumulated kernels

      list<Spectrum>::const_iterator itr;

      unsigned maxKernelSize = 0;
      long new_maxKernelSize = 0;

      // Primitive way to change Doug's code to work also with 64-bit machines
      // without using maximize not changing all the variables
      //      for ( itr = kernels.begin(); itr != kernels.end(); ++itr )
      //	 maximize( maxKernelSize, itr->size() );

      for ( itr = kernels.begin(); itr != kernels.end(); ++itr )
      {
    	  new_maxKernelSize = max ( static_cast<long>(maxKernelSize), static_cast<long>(itr->size()) );
    	  maxKernelSize = static_cast<unsigned>(new_maxKernelSize);
      }
      unsigned size = dist.size() + maxKernelSize;
      unsigned padding = (1 + maxKernelSize) / 2;

      struct Convolver
      	  convolver( 1 << (int) ceil(
				   static_cast<long double>(std::log(static_cast<long double>(size+padding))) /
				   static_cast<long double>(std::log(static_cast<long double>(2))) ) );

      vector<float> curve( convolver.size() );
      shift( curve, dist, (int)size/2 - (int)dist.size()/2 );
      convolver.load( curve );

      for ( itr = kernels.begin(); itr != kernels.end(); ++itr )
    	  if ( itr->size() ) convolver.load( wrap(curve,*itr) );

      dist.resize( size );
      dist.assign( convolver().begin(), convolver().begin() + dist.size() );

    }
  else
    dist = caldsGR->scatteringDistribution(3,dBeta,
					   alpha,order,energy,true,false);

  shift( rf, dist, betaChannelGrid.cell(energy_m)-betaChannelGrid.min-(int)dist.size()/2 );

  double effective_area = eaCurve[-order]->area(1000.*eLow,1000.*eHigh);


  //for ( unsigned i = 0; i < betaGrid.length(); ++i ) rf[i] *= effective_area;
  for ( unsigned i = 0; i < betaChannelGrid.length(); ++i ) rf[i] *= effective_area;   //AI: Wavelength Scale. betaChannelGrid

}

/* --- calibrations -------------------------------------------------------- */

float empiricalCrossLA( float energy, float dpsi, float xdsp0, float xdsp1 )
{
  float k = (energy * 10800) / (0.678 * M_PI);
  return (atan(k*(xdsp1+dpsi)) - atan(k*(xdsp0+dpsi))) / M_PI;
}

double calBeta( double phi /* arcmin */, int order, double energy /* eV */ )
{
  return caldsGR->beta( caldsGR->alpha(-60*phi,0), order, energy );
}

float calEnergy( float beta, int order, float alpha )
{
  return caldsGR->energy(alpha,beta,order);
}

//AI: Wavelength scale. Start
float calLambdaEnergy( float lambda )
{
  return caldsGR->energy(lambda);
}
//AI: Wavelength scale. End

float calAlpha0()
{
  return caldsGR->alpha0();
}

void ccd_qe( float qe[10], float energy /* keV */ )
{
  qe[0] = 0;
  for ( unsigned i = 1; i <= 9; ++i )
    qe[i] = ccdqeCurve[i-1].efficiency( energy * 1000 );

}

Spectrum customFigureDistribution( double betaBinWidth,
				   int order, double energy )
{
  struct Grid grid( betaBinWidth, calBeta(0,order,energy) );

  double betaMin = calBeta( customFigure.front().phi0, order, energy );
  double betaMax = calBeta( customFigure.back().phi1,  order, energy );

  int width = max( grid.cell(betaMax), -grid.cell(betaMin) );

  struct Spectrum dist( 1 + 2*width );
  grid.origin -= width * grid.stride;

  vector<AngularDistBin>::const_iterator itr;
  for ( itr = customFigure.begin(); itr != customFigure.end(); ++itr ) {
    float idx0 = 0.5 + grid.tick( calBeta(itr->phi0,order,energy) );
    float idx1 = 0.5 + grid.tick( calBeta(itr->phi1,order,energy) );
    rebin( dist, idx0, idx1, itr->value );
  }

  return dist;
}

/* --- initialize/finalize ------------------------------------------------- */

void calibInit( const GridSpan &_betaGrid, const GridSpan &_betaChannelGrid,
		const GridSpan &_betaORIChannelGrid,  int order)
{
  betaGrid = _betaGrid;
  betaChannelGrid = _betaChannelGrid;
  betaORIChannelGrid = _betaORIChannelGrid;

  caldsGR = 0; caldsGR = calServer.getAtom(caldsGR);
  caldsEA = 0; caldsEA = calServer.getAtom(caldsEA);
  caldsPS = 0; caldsPS = calServer.getAtom(caldsPS);
  caldsGM = 0; caldsGM = calServer.getAtom(caldsGM);

  unsigned i;
  for ( i = 0; i < 6; ++i ) eaCurve[i] = 0;

  for ( i = 0; i < 9; ++i ) {
    struct CcdQuantumDataServer *qds=0;
    calServer.state()->ccd()->set(i+1);
    qds = calServer.getAtom(qds);
    //AI: Applying the CCD correction factor per ccd and order
    errstr << "Retrieving CCD Correction factor for CCD  " << i+1
	   << Sas::message(Msg::App, Msg::Verbose);
    qds->ccdCorrFactors(order);
    ccdqeCurve[i] = qds->grade(1);
    calServer.releaseAtom(qds);
  }

  //AI: In case we are dealing with lambda grid, this calculation is overridden with
  //the function below.
  vector<double> realBeta(betaChannelGrid.length()); //AI: Wavelength scale. Start
  for ( i=0; i < betaChannelGrid.length(); ++i ) //AI: Wavelength scale. Start
	  realBeta[i] = betaChannelGrid[i+betaChannelGrid.min]; //AI: Wavelength scale. Start

  //vignetting = caldsGR->selfVignettingEfficiency(beta);
  vignetting = caldsGR->selfVignettingEfficiency(realBeta); //AI: Wavelength scale.

}


//Override the vignetting values for lambda energy grid!!
void calVignetting( const GridSpan &betaChannelGrid,  float alpha)
{

  vector<double> realBeta(betaChannelGrid.length()); //AI: Wavelength scale. Start
  for ( unsigned int i=0; i < betaChannelGrid.length(); ++i )//AI: Wavelength scale. Start
  {
	   double lambda = betaChannelGrid[i+betaChannelGrid.min]; //AI: Wavelength scale. Start
	   realBeta[i] = caldsGR->lambda2beta(alpha, -1, lambda);
  }
  vignetting = caldsGR->selfVignettingEfficiency(realBeta); //AI: Wavelength scale.

}


void resetSourcePosition( float _theta, float _phi )
{
  theta = _theta;
  phi = _phi;

  /*
   * INFO
   */
  if ( booleanParameter("dyneffareacorr"))
    errstr << "Using Effective Area Curve based on chebyshev polinomials "
	   << Sas::message(Msg::App, Msg::Verbose);
  else
    errstr << "Using old Effective Area Curve Correction"
	   << Sas::message(Msg::App, Msg::Verbose);


  for ( unsigned i = 0; i < 6; ++i )
    {
      delete eaCurve[i];
      EffectiveAreaCurve effAreaCurve;

      //AI: Call to the dynamic effective area correction (related to the
      // Chebyshev polinomials.
      if ( booleanParameter("dyneffareacorr"))
    	  effAreaCurve = caldsEA->realisticEffectiveAreaCurve(-i,theta,phi);
      else
    	  effAreaCurve = caldsEA->intrinsicEffectiveAreaCurve(-i,theta,phi);

      if ( booleanParameter("withrectification"))
    	  effAreaCurve =  caldsEA->effectiveAreaRectification(effAreaCurve);

      eaCurve[i] = new EffectiveAreaCurve(effAreaCurve);
    }

}

void calibFree()
{
  for ( unsigned i = 0; i < 6; ++i ) delete eaCurve[i];
  calServer.releaseAtom(caldsGM);
  calServer.releaseAtom(caldsPS);
  calServer.releaseAtom(caldsEA);
  calServer.releaseAtom(caldsGR);
}

/* --- utility ------------------------------------------------------------- */

vector<float>& wrap( vector<float> &curve, const Spectrum &dist )
{
  vector<double>::const_iterator center, begin, end, itr;

  if ( dist.size() > curve.size() ) {
    begin = dist.begin() + (dist.size() - curve.size())/2;
    end = begin + curve.size();
  }
  else {
    begin = dist.begin();
    end = dist.end();
  }

  center = dist.begin() + dist.size()/2;
  if ( center < begin )
    center = begin;
  else if ( center > end )
    center = end;

  float *ptr = &curve.front();
  for ( itr = center; itr != end; *ptr++ = *itr++ ) ;

  unsigned count = curve.size() - (end-begin);
  memset( ptr, 0, count * sizeof(float) );
  ptr += count;

  for ( itr = begin; itr != center; *ptr++ = *itr++ ) ;

  return curve;
}

void shift( vector<float> &curve, const Spectrum &dist, int offset )
{

  float *ptr = &curve.front();

  int begin = max( offset, 0 );
  int end = min( (int)dist.size() + offset, (int)curve.size() );

  if ( end < begin ) {
    memset( ptr, 0, curve.size() * sizeof(float) );
    return;
  }

  if ( begin > 0 ) {
    memset( ptr, 0, begin * sizeof(float) );
    ptr += begin;
  }

  vector<double>::const_iterator itr, endI = dist.begin() + end - offset;
  for ( itr = dist.begin() + begin - offset; itr != endI; *ptr++ = *itr++ ) ;

  if ( (int)curve.size() > end )
    memset( ptr, 0, (curve.size() - end) * sizeof(float) );
}

void rebin( struct Spectrum &vec, float min, float max, float val )
{
  if ( max < min ) swap(min,max);

  if ( min >= vec.size() || max <= 0 ) return;

  if ( min < 0 ) {
    val *= 1 + min/(max - min);
    min = 0;
  }

  if ( max > vec.size() ) {
    val *= 1 - (max - vec.size())/(max - min);
    max = vec.size();
  }

  if ( floor(max) > floor(min) ) {
    float portion = val / (max - min);
    if ( ceil(min) > min ) {
      vec[int(floor(min))] += portion * (ceil(min) - min);
      min = ceil(min);
    }
    while ( floor(max) > floor(min) ) {
      vec[int(floor(min))] += portion;
      min += 1;
    }
    vec[int(floor(min))] += portion * (max - min);
  }
  else
    vec[int(floor(min))] += val;
}
